import { JWT_ALGORITHM, JWT_EXPIRATION, JWT_ISSUER } from '$env/static/private';
import { readFile } from 'fs/promises';
import { SignJWT, importPKCS8, importSPKI, jwtVerify } from 'jose';

const privateKeyFile = await readFile('private/jwt.private.pem', { encoding: 'utf-8' });
const publicKeyFile = await readFile('private/jwt.public.pem', { encoding: 'utf-8' });
const privateKey = await importPKCS8(privateKeyFile, JWT_ALGORITHM);
const publicKey = await importSPKI(publicKeyFile, JWT_ALGORITHM);

/**
 * Generate JWTs using the following commands:
 * Private: openssl genpkey -out jwt.private.pem -algorithm RSA -pkeyopt rsa_keygen_bits:2048
 * Public:  openssl rsa -in jwt.private.pem -pubout -outform PEM -out jwt.public.pem
 */
export class JWT {
	static privateKey = privateKey;
	static publicKey = publicKey;

	static async issue(claims: Record<string, unknown>, subject: string, audience?: string) {
		const sign = new SignJWT(claims)
			.setProtectedHeader({ alg: JWT_ALGORITHM })
			.setIssuedAt()
			.setSubject(subject)
			.setExpirationTime(JWT_EXPIRATION)
			.setIssuer(JWT_ISSUER);

		if (audience) {
			sign.setAudience(audience);
		}

		return sign.sign(JWT.privateKey);
	}

	static async verify(token: string, subject?: string, audience?: string) {
		const { payload } = await jwtVerify(token, JWT.publicKey, {
			issuer: JWT_ISSUER,
			subject,
			audience
		});
		return payload;
	}
}
