import { PUBLIC_URL, PUBLIC_SITE_NAME } from '$env/static/public';
import { CryptoUtils } from '$lib/server/crypto-utils';
import {
	db,
	oauth2Client,
	oauth2ClientAuthorization,
	oauth2ClientManager,
	oauth2ClientUrl,
	privilege,
	user,
	userPrivilegesPrivilege,
	type OAuth2Client,
	type OAuth2ClientUrl,
	type User
} from '$lib/server/drizzle';
import { Emails, OAuth2InvitationEmail } from '$lib/server/email';
import { Uploads } from '$lib/server/upload';
import { UserTokens } from '$lib/server/users';
import type { PaginationMeta } from '$lib/types';
import { and, count, eq, like, or, sql } from 'drizzle-orm';

export enum OAuth2ClientURLType {
	REDIRECT_URI = 'redirect_uri',
	TERMS = 'terms',
	PRIVACY = 'privacy',
	WEBSITE = 'website'
}

export interface OAuth2ClientAdminListItem
	extends PartialK<OAuth2Client, 'client_secret' | 'ownerId'> {
	isOwner?: boolean;
	ownerInfo?: { uuid: string; name: string };
	urls: Omit<OAuth2ClientUrl, 'created_at' | 'updated_at' | 'clientId'>[];
}

export interface OAuth2AuthorizedUser {
	uuid: string;
	name: string;
	current: boolean;
	privileges: { id: number; name: string }[];
}

export interface OAuth2ManagerUser {
	id: number;
	email: string;
}

export class OAuth2Clients {
	public static availableGrantTypes = [
		'authorization_code',
		'client_credentials',
		'refresh_token',
		'id_token',
		'implicit'
	];

	public static availableScopes = [
		'profile',
		'picture',
		'email',
		'privileges',
		'management',
		'account',
		'openid'
	];

	public static userSetScopes = [
		'profile',
		'picture',
		'email',
		'privileges',
		'management',
		'openid'
	];
	public static userSetGrants = [
		'authorization_code',
		'client_credentials',
		'refresh_token',
		'id_token'
	];

	public static describedScopes = ['email', 'picture', 'account'];
	public static alwaysPresentScopes = ['profile'];

	public static availableUrlTypes: OAuth2ClientURLType[] = Object.values(OAuth2ClientURLType);

	static async fetchById(id: string | number) {
		const [client] = await db
			.select()
			.from(oauth2Client)
			.where(typeof id === 'string' ? eq(oauth2Client.client_id, id) : eq(oauth2Client.id, id))
			.limit(1);
		return client;
	}

	static async getRedirectUrls(
		id: string,
		type: OAuth2ClientURLType = OAuth2ClientURLType.REDIRECT_URI
	) {
		return await db
			.select()
			.from(oauth2ClientUrl)
			.innerJoin(oauth2Client, eq(oauth2ClientUrl.clientId, oauth2Client.id))
			.where(and(eq(oauth2Client.client_id, id), eq(oauth2ClientUrl.type, type)));
	}

	static async getClientUrls(client: OAuth2Client) {
		return await db
			.select()
			.from(oauth2ClientUrl)
			.where(and(eq(oauth2ClientUrl.clientId, client.id)));
	}

	static async checkRedirectUri(client: OAuth2Client, url: string) {
		return !!(
			await db
				.select()
				.from(oauth2ClientUrl)
				.innerJoin(oauth2Client, eq(oauth2ClientUrl.clientId, oauth2Client.id))
				.where(
					and(
						eq(oauth2Client.client_id, client.client_id),
						eq(oauth2ClientUrl.type, OAuth2ClientURLType.REDIRECT_URI),
						eq(oauth2ClientUrl.url, url)
					)
				)
		)?.length;
	}

	static async getAuthorizedUsers(client: OAuth2Client, userUuid?: string) {
		const junkList = await db
			.select()
			.from(oauth2ClientAuthorization)
			.innerJoin(user, eq(user.id, oauth2ClientAuthorization.userId))
			.leftJoin(userPrivilegesPrivilege, eq(userPrivilegesPrivilege.userId, user.id))
			.leftJoin(
				privilege,
				and(
					eq(userPrivilegesPrivilege.privilegeId, privilege.id),
					eq(privilege.clientId, client.id)
				)
			)
			.where(
				and(
					eq(oauth2ClientAuthorization.clientId, client.id),
					userUuid ? eq(user.uuid, userUuid) : undefined
				)
			);

		const list = junkList.reduce<OAuth2AuthorizedUser[]>((accum, dbo) => {
			let user = accum.find((entry) => entry.uuid === dbo.user.uuid);
			if (!user) {
				user = {
					uuid: dbo.user.uuid,
					name: dbo.user.display_name,
					current: false,
					privileges: []
				};
				accum.push(user);
			}

			if (dbo.privilege && !user.privileges.some(({ id }) => id === dbo.privilege?.id)) {
				user.privileges.push({ id: dbo.privilege.id, name: dbo.privilege.name });
			}

			if (dbo.o_auth2_client_authorization?.current === 1) {
				user.current = true;
			}

			return accum;
		}, []);

		return list;
	}

	static checkSecret(client: OAuth2Client, secret: string) {
		return client.client_secret === secret;
	}

	static checkGrantType(client: OAuth2Client, grant: string) {
		return client.grants.split(' ').includes(grant);
	}

	static transformScope(scope: string | string[]): string[] {
		return Array.isArray(scope) ? scope : OAuth2Clients.splitScope(scope);
	}

	static checkScope(client: OAuth2Client, scope: string[]): boolean {
		return scope.every((one) => one === 'profile' || client.scope?.includes(one));
	}

	static splitScope(scope: string | string[]): string[] {
		if (!scope) {
			return [];
		}

		if (Array.isArray(scope)) {
			return scope;
		}

		return scope.includes(',') ? scope.split(',').map((item) => item.trim()) : scope.split(' ');
	}

	static joinScope(scope: string[]): string {
		return scope.join(' ');
	}

	static async getClientByAdminUser(
		subject: User,
		filters?: {
			clientId?: string;
			filter?: string;
			limit?: number;
			offset?: number;
			omitSecret?: boolean;
			listAll?: boolean;
		}
	) {
		const filterText = `%${filters?.filter?.toLowerCase()}%`;
		const limit = filters?.limit || 20;
		const allowedClients = db
			.select({ id: oauth2Client.id })
			.from(oauth2Client)
			.leftJoin(oauth2ClientManager, eq(oauth2ClientManager.clientId, oauth2Client.id))
			.where(
				and(
					!filters?.listAll
						? or(eq(oauth2Client.ownerId, subject.id), eq(oauth2ClientManager.userId, subject.id))
						: undefined,
					filters?.clientId ? eq(oauth2Client.client_id, filters.clientId) : undefined,
					filters?.filter
						? or(
								sql`lower(${oauth2Client.title}) like ${filterText}`,
								like(oauth2Client.client_id, filterText)
							)
						: undefined
				)
			)
			.groupBy(oauth2Client.id)
			.limit(limit)
			.offset(filters?.offset || 0)
			.as('allowedClients');

		const [{ rowCount }] = await db
			.select({
				rowCount: count(oauth2Client.id).mapWith(Number)
			})
			.from(allowedClients)
			.innerJoin(oauth2Client, eq(allowedClients.id, oauth2Client.id));

		const junkList = await db
			.select({
				o_auth2_client: oauth2Client,
				o_auth2_client_url: oauth2ClientUrl,
				user: user
			})
			.from(allowedClients)
			.innerJoin(oauth2Client, eq(allowedClients.id, oauth2Client.id))
			.leftJoin(oauth2ClientUrl, eq(oauth2ClientUrl.clientId, oauth2Client.id))
			.leftJoin(user, eq(oauth2Client.ownerId, user.id));

		const list = junkList.reduce<OAuth2ClientAdminListItem[]>((accum, dbo) => {
			let client = accum.find(({ id }) => id === dbo.o_auth2_client.id);
			if (!client) {
				client = {
					...dbo.o_auth2_client,
					isOwner: dbo.o_auth2_client.ownerId === subject.id,
					urls: []
				};
				accum.push(client);
			}

			if (dbo.o_auth2_client_url) {
				if (!client.urls.some(({ id }) => dbo.o_auth2_client_url?.id === id)) {
					client.urls.push({
						...dbo.o_auth2_client_url,
						clientId: undefined,
						created_at: undefined,
						updated_at: undefined
					} as OAuth2ClientAdminListItem['urls'][0]);
				}
			}

			if (dbo.user) {
				client.ownerInfo = {
					uuid: dbo.user.uuid,
					name: dbo.user.display_name
				};
			}

			if (filters?.omitSecret) {
				delete client.client_secret;
			}

			return accum;
		}, []);

		const meta: PaginationMeta = {
			rowCount,
			pageSize: limit,
			pageCount: Math.ceil(rowCount / limit)
		};

		return {
			list,
			meta
		};
	}

	static async createClient(subject: User, title: string, redirect: string, description?: string) {
		const uid = CryptoUtils.createUUID();
		const secret = CryptoUtils.generateSecret();

		const [retval] = await db.insert(oauth2Client).values({
			title,
			description,
			client_id: uid,
			client_secret: secret,
			grants: 'authorization_code',
			scope: 'profile',
			ownerId: subject.id,
			created_at: new Date(),
			activated: 1,
			verified: 0
		});

		await db.insert(oauth2ClientUrl).values({
			type: 'redirect_uri',
			url: redirect,
			clientId: retval.insertId
		});

		return uid;
	}

	static async deleteClient(client: OAuth2Client) {
		if (client.pictureId) {
			await Uploads.removeClientAvatar(client);
		}
		await db.delete(privilege).where(eq(privilege.clientId, client.id));
		await db.delete(oauth2Client).where(eq(oauth2Client.id, client.id));
	}

	static async deleteUrl(client: OAuth2Client, urlId: number) {
		await db
			.delete(oauth2ClientUrl)
			.where(and(eq(oauth2ClientUrl.clientId, client.id), eq(oauth2ClientUrl.id, urlId)));
	}

	static async addUrl(client: OAuth2Client, type: OAuth2ClientURLType, url: string) {
		await db.insert(oauth2ClientUrl).values({
			type,
			url,
			clientId: client.id
		});
	}

	static async deletePrivilege(client: OAuth2Client, privilegeId: number) {
		await db
			.delete(privilege)
			.where(and(eq(privilege.clientId, client.id), eq(privilege.id, privilegeId)));
	}

	static async addPrivilege(client: OAuth2Client, name: string) {
		const realName = `${client.client_id.split('-')[0]}:${name}`;
		await db.insert(privilege).values({
			name: realName,
			clientId: client.id
		});
	}

	static async update(client: OAuth2Client, body: Partial<OAuth2Client>) {
		await db.update(oauth2Client).set(body).where(eq(oauth2Client.id, client.id));
	}

	static async getManagers(client: OAuth2Client) {
		return await db
			.select({ id: oauth2ClientManager.id, email: user.email })
			.from(oauth2ClientManager)
			.innerJoin(user, eq(user.id, oauth2ClientManager.userId))
			.where(eq(oauth2ClientManager.clientId, client.id));
	}

	static async addManager(client: OAuth2Client, actor: User, subject: User) {
		await db.insert(oauth2ClientManager).values({
			clientId: client.id,
			userId: subject.id,
			issuerId: actor.id
		});
	}

	static async removeManager(client: OAuth2Client, managerId: number) {
		await db
			.delete(oauth2ClientManager)
			.where(
				and(eq(oauth2ClientManager.clientId, client.id), eq(oauth2ClientManager.id, managerId))
			);
	}

	static async sendManagerInvitationEmail(client: OAuth2Client, actor: User, email: string) {
		const token = await UserTokens.create(
			'invite',
			new Date(Date.now() + 3600 * 1000),
			actor.id,
			undefined,
			`clientmanager=${client.client_id}`
		);
		const params = new URLSearchParams({ token: token.token });
		const content = OAuth2InvitationEmail(
			actor.display_name,
			client.title,
			`${PUBLIC_URL}/ssoadmin/oauth2/invite?${params.toString()}`
		);

		// TODO: logging
		try {
			await Emails.getSender().sendTemplate(
				email,
				`You have been invited to manage "${client.title}" on ${PUBLIC_SITE_NAME}`,
				content
			);
		} catch {
			await UserTokens.remove(token);
		}
	}

	static async authorizeClientInfo(client: OAuth2Client, scope: string[]) {
		const links = await OAuth2Clients.getClientUrls(client);
		const filteredLinks = links
			.filter((link) => link.type !== 'redirect_uri')
			.map(({ url, type }) => ({
				url,
				type
			}));
		const allowedScopes = [...OAuth2Clients.alwaysPresentScopes];
		const disallowedScopes: string[] = [];

		OAuth2Clients.describedScopes.forEach((item) => {
			if (scope.includes(item)) {
				allowedScopes.push(item);
			} else {
				disallowedScopes.push(item);
			}
		});

		if (scope.includes('management')) {
			allowedScopes.push('management');
		}

		return {
			links: filteredLinks,
			client_id: client.client_id,
			title: client.title,
			description: client.description,
			grants: client.grants,
			allowedScopes,
			disallowedScopes
		};
	}
}
