import {
	db,
	oauth2Client,
	oauth2Token,
	type OAuth2Client,
	type OAuth2Token,
	type User
} from '$lib/server/drizzle';
import { and, eq, sql } from 'drizzle-orm';
import { OAuth2Clients } from './client';
import { Users } from '$lib/server/users';
import { CryptoUtils } from '$lib/server/crypto-utils';

export type CodeChallengeMethod = 'plain' | 'S256';

export enum OAuth2TokenType {
	CODE = 'code',
	ACCESS_TOKEN = 'access_token',
	REFRESH_TOKEN = 'refresh_token'
}

export interface OAuth2Code extends OAuth2Token {
	code_challenge?: string;
	code_challenge_method?: CodeChallengeMethod;
	clientIdPub: string;
}

export interface OAuth2AccessToken extends OAuth2Token {
	clientIdPub: string;
}

export interface OAuth2RefreshToken extends OAuth2Token {
	clientIdPub: string;
}

export class OAuth2Tokens {
	static codeTtl = 3600;
	static tokenTtl = 604800;
	static refreshTtl = 3.154e7;
	static challengeMethods: CodeChallengeMethod[] = ['plain', 'S256'];

	static async insert(
		token: string,
		type: OAuth2TokenType,
		client: OAuth2Client,
		scope: string,
		expiry: Date,
		user?: User,
		nonce?: string,
		pcke?: string
	) {
		const [retval] = await db.insert(oauth2Token).values({
			token,
			type,
			scope,
			expires_at: expiry,
			clientId: client.id,
			userId: user?.id,
			nonce,
			pcke
		});

		const [newToken] = await db
			.select()
			.from(oauth2Token)
			.where(eq(oauth2Token.id, retval.insertId));

		return newToken;
	}

	static async fetchByToken(token: string, type: OAuth2TokenType) {
		const [retval] = await db
			.select()
			.from(oauth2Token)
			.where(and(eq(oauth2Token.token, token), eq(oauth2Token.type, type)));
		return retval;
	}

	static async fetchByUserIdClientId(userId: number, clientId: string, type: OAuth2TokenType) {
		const [retval] = await db
			.select()
			.from(oauth2Token)
			.innerJoin(oauth2Client, eq(oauth2Token.clientId, oauth2Client.id))
			.where(
				and(
					eq(oauth2Client.client_id, clientId),
					eq(oauth2Token.type, type),
					eq(oauth2Token.userId, userId)
				)
			);
		return retval;
	}

	static async wipeClientTokens(client: OAuth2Client, user?: User) {
		await db
			.delete(oauth2Token)
			.where(
				and(eq(oauth2Token.clientId, client.id), user ? eq(oauth2Token.userId, user.id) : undefined)
			);
	}

	static async wipeUserTokens(user: User) {
		await db.delete(oauth2Token).where(eq(oauth2Token.userId, user.id));
	}

	static async wipeExpiredTokens() {
		await db.execute(sql`DELETE FROM ${oauth2Token} WHERE ${oauth2Token.expires_at} < NOW()`);
	}

	static async remove(token: OAuth2Token) {
		await db.delete(oauth2Token).where(eq(oauth2Token.id, token.id));
	}
}

export class OAuth2Codes {
	static async create(
		userId: number,
		clientId: string,
		scope: string | string[],
		ttl: number,
		nonce?: string,
		codeChallenge?: string,
		codeChallengeMethod?: CodeChallengeMethod
	) {
		const client = await OAuth2Clients.fetchById(clientId);
		const user = await Users.getById(userId);
		const accessToken = CryptoUtils.generateString(64);

		const scopes = (!Array.isArray(scope) ? OAuth2Clients.splitScope(scope) : scope).join(' ');

		const expiresAt = new Date(Date.now() + ttl * 1000);
		const pcke =
			codeChallenge && codeChallengeMethod
				? `${OAuth2Tokens.challengeMethods.indexOf(codeChallengeMethod)}:${codeChallenge}`
				: undefined;

		await OAuth2Tokens.insert(
			accessToken,
			OAuth2TokenType.CODE,
			client,
			scopes,
			expiresAt,
			user,
			nonce,
			pcke
		);

		return accessToken;
	}

	static async fetchByCode(code: string | OAuth2Token): Promise<OAuth2Code | undefined> {
		const findBy = typeof code === 'string' ? code : code.token;
		const find = await OAuth2Tokens.fetchByToken(findBy, OAuth2TokenType.CODE);
		if (!find) {
			return undefined;
		}

		let codeChallenge: string | undefined;
		let codeChallengeMethod: CodeChallengeMethod | undefined;
		if (find.pcke) {
			codeChallengeMethod = OAuth2Tokens.challengeMethods[Number(find.pcke.substring(0, 1))];
			codeChallenge = find.pcke.substring(2);
		}

		const client = await OAuth2Clients.fetchById(find.clientId as number);
		return {
			...find,
			clientIdPub: client.client_id,
			code_challenge: codeChallenge,
			code_challenge_method: codeChallengeMethod
		};
	}

	static async removeByCode(code: string | OAuth2Token): Promise<boolean> {
		const findBy = typeof code === 'string' ? code : code.token;
		const find = await OAuth2Tokens.fetchByToken(findBy, OAuth2TokenType.CODE);
		await OAuth2Tokens.remove(find);
		return true;
	}

	static checkTTL(code: OAuth2Code): boolean {
		return new Date(code.expires_at).getTime() > Date.now();
	}

	static getCodeChallenge(code: OAuth2Code) {
		return {
			method: code.code_challenge_method,
			challenge: code.code_challenge
		};
	}
}

export class OAuth2AccessTokens {
	static async create(
		userId: number | null,
		clientId: string,
		scope: string | string[],
		ttl: number
	) {
		const client = await OAuth2Clients.fetchById(clientId);
		const user = userId != null ? await Users.getById(userId) : undefined;
		const accessToken = CryptoUtils.generateString(64);

		const scopes = (!Array.isArray(scope) ? OAuth2Clients.splitScope(scope) : scope).join(' ');

		const expiresAt = new Date(Date.now() + ttl * 1000);

		await OAuth2Tokens.insert(
			accessToken,
			OAuth2TokenType.ACCESS_TOKEN,
			client,
			scopes,
			expiresAt,
			user
		);

		return accessToken;
	}

	static async fetchByToken(token: string | OAuth2Token): Promise<OAuth2AccessToken | undefined> {
		const findBy = typeof token === 'string' ? token : token.token;
		const find = await OAuth2Tokens.fetchByToken(findBy, OAuth2TokenType.ACCESS_TOKEN);
		if (!find) {
			return undefined;
		}

		const client = await OAuth2Clients.fetchById(find.clientId as number);
		return {
			...find,
			clientIdPub: client.client_id
		};
	}

	static checkTTL(token: OAuth2AccessToken): boolean {
		return new Date() < new Date(token.expires_at);
	}

	static getTTL(token: OAuth2AccessToken): number {
		return new Date(token.expires_at).getTime() - Date.now();
	}

	static async fetchByUserIdClientId(
		userId: number,
		clientId: string
	): Promise<OAuth2AccessToken | undefined> {
		const find = await OAuth2Tokens.fetchByUserIdClientId(
			userId,
			clientId,
			OAuth2TokenType.ACCESS_TOKEN
		);
		if (!find) {
			return undefined;
		}

		return {
			...find.o_auth2_token,
			clientIdPub: find.o_auth2_client.client_id
		};
	}
}

export class OAuth2RefreshTokens {
	static async create(userId: number, clientId: string, scope: string | string[]) {
		const client = await OAuth2Clients.fetchById(clientId);
		const user = await Users.getById(userId);
		const accessToken = CryptoUtils.generateString(64);

		const scopes = (!Array.isArray(scope) ? OAuth2Clients.splitScope(scope) : scope).join(' ');

		const expiresAt = new Date(Date.now() + OAuth2Tokens.refreshTtl * 1000);

		await OAuth2Tokens.insert(
			accessToken,
			OAuth2TokenType.REFRESH_TOKEN,
			client,
			scopes,
			expiresAt,
			user
		);

		return accessToken;
	}

	static async fetchByToken(token: string | OAuth2Token): Promise<OAuth2RefreshToken | undefined> {
		const findBy = typeof token === 'string' ? token : token.token;
		const find = await OAuth2Tokens.fetchByToken(findBy, OAuth2TokenType.REFRESH_TOKEN);
		if (!find) {
			return undefined;
		}

		const client = await OAuth2Clients.fetchById(find.clientId as number);
		return {
			...find,
			clientIdPub: client.client_id
		};
	}

	static async removeByRefreshToken(token: string): Promise<boolean> {
		const find = await OAuth2Tokens.fetchByToken(token, OAuth2TokenType.REFRESH_TOKEN);

		await OAuth2Tokens.remove(find);

		return true;
	}
}
