import { count, eq, like, or, sql } from 'drizzle-orm';
import {
	db,
	privilege,
	user,
	userPrivilegesPrivilege,
	type Privilege,
	type User
} from '../drizzle';
import type { Paginated, PaginationMeta } from '$lib/types';
import type { RequiredPrivileges } from '$lib/utils';
import { Users } from '.';
import { error } from '@sveltejs/kit';
import { AdminUtils } from '../admin-utils';

export interface AdminUserListItem extends Omit<User, 'password'> {
	privileges: Privilege[];
}

export class UsersAdmin {
	static mergeUserResponse(
		junkList: {
			user: User;
			user_privileges_privilege?: typeof userPrivilegesPrivilege.$inferSelect | null;
			privilege?: Privilege | null;
		}[]
	) {
		return junkList.reduce<AdminUserListItem[]>((accum, dbe) => {
			let user = accum.find((entry) => entry.id === dbe.user.id);
			if (!user) {
				user = { ...dbe.user, password: undefined, privileges: [] } as AdminUserListItem;
				accum.push(user);
			}

			// Individual privilege
			if (dbe.user_privileges_privilege && dbe.privilege) {
				if (
					!user.privileges.some((priv) => priv.id === dbe.user_privileges_privilege?.privilegeId)
				) {
					user.privileges.push(dbe.privilege);
				}
			}

			return accum;
		}, []);
	}

	static async getAllUsers({
		filter,
		offset = 0,
		limit = 20
	}: {
		filter?: string;
		offset?: number;
		limit?: number;
	}) {
		const subfilter = `%${filter}%`;
		const searchExpression = filter
			? or(
					like(user.uuid, subfilter),
					sql`lower(${user.username}) like ${subfilter}`,
					sql`lower(${user.display_name}) like ${subfilter}`,
					sql`lower(${user.email}) like ${subfilter}`
				)
			: undefined;

		const [{ rowCount }] = await db
			.select({ rowCount: count(user.id).mapWith(Number) })
			.from(user)
			.where(searchExpression);

		const junkList = await db
			.select()
			.from(user)
			.leftJoin(userPrivilegesPrivilege, eq(userPrivilegesPrivilege.userId, user.id))
			.leftJoin(privilege, eq(userPrivilegesPrivilege.privilegeId, privilege.id))
			.where(searchExpression)
			.limit(limit)
			.offset(offset);

		const meta: PaginationMeta = {
			rowCount,
			pageSize: limit,
			pageCount: Math.ceil(rowCount / limit)
		};

		const list = UsersAdmin.mergeUserResponse(junkList);

		return <Paginated<AdminUserListItem>>{
			list,
			meta
		};
	}

	static async getUserDetails(uuid: string) {
		const junkList = await db
			.select()
			.from(user)
			.leftJoin(userPrivilegesPrivilege, eq(userPrivilegesPrivilege.userId, user.id))
			.leftJoin(privilege, eq(userPrivilegesPrivilege.privilegeId, privilege.id))
			.where(eq(user.uuid, uuid));
		const [userInfo] = UsersAdmin.mergeUserResponse(junkList);
		return userInfo;
	}

	static async getActionUser(locals: App.Locals, privileges: RequiredPrivileges) {
		const userSession = locals.session.data?.user;
		const currentUser = await Users.getBySession(userSession);
		if (!userSession || !currentUser) {
			return error(403);
		}

		userSession.privileges = await Users.getUserPrivileges(currentUser);
		AdminUtils.checkPrivileges(userSession, privileges);

		return { currentUser, userSession };
	}
}
