import bcrypt from 'bcryptjs';
import { and, eq, inArray, isNull, or, sql } from 'drizzle-orm';
import { db, privilege, user, userPrivilegesPrivilege, type User } from '../drizzle';
import type { UserSession } from './types';
import { redirect } from '@sveltejs/kit';
import { CryptoUtils } from '../crypto-utils';
import { EMAIL_ENABLED } from '$env/static/private';
import { Emails, ForgotPasswordEmail, InvitationEmail, RegistrationEmail } from '../email';
import { PUBLIC_SITE_NAME, PUBLIC_URL } from '$env/static/public';
import { UserTokens } from './tokens';

export class Users {
	static async getById(id: number): Promise<User | undefined> {
		const [result] = await db
			.select()
			.from(user)
			.where(and(eq(user.id, id), eq(user.activated, 1)))
			.limit(1);
		return result;
	}

	static async getByUuid(uuid: string, activatedCheck = true): Promise<User | undefined> {
		const [result] = await db
			.select()
			.from(user)
			.where(and(eq(user.uuid, uuid), activatedCheck ? eq(user.activated, 1) : undefined))
			.limit(1);
		return result;
	}

	static async getByLogin(login: string): Promise<User | undefined> {
		const [result] = await db
			.select()
			.from(user)
			.where(and(or(eq(user.email, login), eq(user.username, login)), eq(user.activated, 1)))
			.limit(1);
		return result;
	}

	static async getBySession(session?: UserSession): Promise<User | undefined> {
		if (!session) return undefined;
		const [result] = await db
			.select()
			.from(user)
			.where(and(eq(user.id, session.uid), eq(user.activated, 1)))
			.limit(1);
		return result;
	}

	static async update(subject: User, fields: Partial<User>) {
		return db.update(user).set(fields).where(eq(user.id, subject.id));
	}

	static async validatePassword(user: User, password: string): Promise<boolean> {
		return bcrypt.compare(password, user.password as string);
	}

	static async hashPassword(password: string): Promise<string> {
		return bcrypt.hash(password, 10);
	}

	static async toSession(user: User): Promise<UserSession> {
		return {
			uid: user.id,
			uuid: user.uuid,
			name: user.display_name,
			username: user.username
		};
	}

	static async readSessionOrRedirect(locals: App.Locals, url: URL) {
		const currentUser = await Users.getBySession(locals.session.data?.user);
		if (!currentUser) {
			await locals.session.destroy();
			return redirect(301, `/login?redirectTo=${encodeURIComponent(url.pathname + url.search)}`);
		}
		return currentUser;
	}

	static async checkRegistration(username: string, email: string) {
		return !(
			await db
				.select({ id: user.id })
				.from(user)
				.where(
					or(
						sql`lower(${user.username}) = ${username.toLowerCase()}`,
						sql`lower(${user.email}) = ${email.toLowerCase()}`
					)
				)
		)?.length;
	}

	static async getActivationToken(token: string) {
		const returnedToken = await UserTokens.getByToken(token, 'activation');
		if (!returnedToken?.userId) return undefined;

		const [userInfo] = await db
			.select()
			.from(user)
			.where(eq(user.id, returnedToken.userId as number));

		return {
			...returnedToken,
			user: userInfo
		};
	}

	static async activateUserBy(token: string, subject: User) {
		await db
			.update(user)
			.set({ activated: 1, activity_at: new Date() })
			.where(eq(user.id, subject.id));
		await UserTokens.remove(token);
	}

	static async register({
		username,
		displayName,
		email,
		password,
		activate = false
	}: {
		username: string;
		displayName: string;
		email: string;
		password: string;
		activate?: boolean;
	}) {
		const passwordHash = await Users.hashPassword(password);
		const [retval] = await db.insert(user).values({
			uuid: CryptoUtils.createUUID(),
			email,
			username,
			password: passwordHash,
			display_name: displayName,
			activated: EMAIL_ENABLED === 'false' ? 1 : Number(activate),
			activity_at: new Date()
		});

		const [newUser] = await db.select().from(user).where(eq(user.id, retval.insertId));

		if (EMAIL_ENABLED !== 'false' && !activate) {
			await Users.sendRegistrationEmail(newUser);
		}

		// TODO: audit log

		return newUser;
	}

	static async sendRegistrationEmail(user: User) {
		const token = await UserTokens.create(
			'activation',
			new Date(Date.now() + 3600 * 1000),
			user.id
		);

		const params = new URLSearchParams({ activate: token.token });
		const content = RegistrationEmail(user.username, `${PUBLIC_URL}/login?${params.toString()}`);

		// TODO: logging
		try {
			await Emails.getSender().sendTemplate(
				user.email,
				`Activate your account on ${PUBLIC_SITE_NAME}`,
				content
			);
		} catch (error) {
			console.error(error);
			await UserTokens.remove(token);
		}
	}

	static async sendPasswordEmail(user: User) {
		const token = await UserTokens.create('password', new Date(Date.now() + 3600 * 1000), user.id);
		const params = new URLSearchParams({ token: token.token });
		const content = ForgotPasswordEmail(
			user.username,
			`${PUBLIC_URL}/login/password?${params.toString()}`
		);

		// TODO: logging
		try {
			await Emails.getSender().sendTemplate(
				user.email,
				`Reset your password on ${PUBLIC_SITE_NAME}`,
				content
			);
		} catch {
			await UserTokens.remove(token);
		}
	}

	static async sendInvitationEmail(email: string) {
		const token = await UserTokens.create(
			'invite',
			new Date(Date.now() + 3600 * 1000),
			undefined,
			undefined,
			`register=${email}`
		);
		const params = new URLSearchParams({ token: token.token });
		const content = InvitationEmail(`${PUBLIC_URL}/register?${params.toString()}`);

		// TODO: logging
		try {
			await Emails.getSender().sendTemplate(
				email,
				`You have been invited to create an account on ${PUBLIC_SITE_NAME}`,
				content
			);
		} catch {
			await UserTokens.remove(token);
		}
	}

	static async getAvailablePrivileges(clientId?: number) {
		return await db
			.select()
			.from(privilege)
			.where(clientId ? eq(privilege.clientId, clientId) : isNull(privilege.clientId));
	}

	static async getUserPrivileges(subject: User, clientId?: number) {
		const list = await db
			.select({
				privilege: privilege.name
			})
			.from(privilege)
			.innerJoin(userPrivilegesPrivilege, eq(privilege.id, userPrivilegesPrivilege.privilegeId))
			.where(
				and(
					eq(userPrivilegesPrivilege.userId, subject.id),
					clientId ? eq(privilege.clientId, clientId) : isNull(privilege.clientId)
				)
			);

		return list.reduce<string[]>(
			(accum, { privilege }) => (!accum.includes(privilege) ? [...accum, privilege] : accum),
			[]
		);
	}

	static async setUserPrivileges(subject: User, privilegeIds: number[], clientId?: number) {
		const current = await db
			.select({
				privilegeId: userPrivilegesPrivilege.privilegeId
			})
			.from(userPrivilegesPrivilege)
			.innerJoin(privilege, eq(privilege.id, userPrivilegesPrivilege.privilegeId))
			.where(
				and(
					eq(userPrivilegesPrivilege.userId, subject.id),
					clientId ? eq(privilege.clientId, clientId) : undefined
				)
			);

		const toRemoveIds = current.reduce<number[]>(
			(list, { privilegeId }) =>
				!privilegeIds.includes(privilegeId) ? [...list, privilegeId] : list,
			[]
		);

		if (toRemoveIds.length) {
			await db
				.delete(userPrivilegesPrivilege)
				.where(
					and(
						eq(userPrivilegesPrivilege.userId, subject.id),
						inArray(userPrivilegesPrivilege.privilegeId, toRemoveIds)
					)
				);
		}

		const toInsertIds = privilegeIds.reduce<number[]>(
			(list, id) => (!current.some(({ privilegeId }) => privilegeId === id) ? [...list, id] : list),
			[]
		);

		if (toInsertIds.length) {
			await db
				.insert(userPrivilegesPrivilege)
				.values(toInsertIds.map((privilegeId) => ({ userId: subject.id, privilegeId })));
		}
	}

	static anonymizeEmail(email: string) {
		const [name, domain] = email.split('@');
		const namePart = `${name.charAt(0)}${''.padStart(name.length - 2, '*')}${name.charAt(name.length - 1)}`;
		return `${namePart}@${domain}`;
	}
}

export * from './totp';
export * from './types';
export * from './tokens';
