import { AdminUtils } from '$lib/server/admin-utils';
import { Changesets } from '$lib/server/changesets.js';
import { CryptoUtils } from '$lib/server/crypto-utils.js';
import type { OAuth2Client, User } from '$lib/server/drizzle';
import { OAuth2ClientURLType, OAuth2Clients } from '$lib/server/oauth2';
import { Uploads } from '$lib/server/upload.js';
import { Users } from '$lib/server/users';
import { UsersAdmin } from '$lib/server/users/admin';
import { hasPrivileges } from '$lib/utils';
import { emailRegex, privilegeRegex } from '$lib/validators.js';
import { error, fail, redirect } from '@sveltejs/kit';

interface AddUrlRequest {
	type: OAuth2ClientURLType;
	url: string;
}

interface UpdateRequest {
	title: string;
	description: string;
	activated?: string;
	verified?: string;
}

interface AddPrivilegeRequest {
	name: string;
}

interface InviteRequest {
	email: string;
}

export const actions = {
	update: async ({ locals, request, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const body = await request.formData();
		const { title, description, activated, verified } = Changesets.take<UpdateRequest>(
			['title', 'description', 'activated', 'verified'],
			body
		);

		if (!!verified && !fullPrivileges) {
			return fail(403, { errors: ['forbidden'] });
		}

		const actuallyVerified = fullPrivileges ? Number(!!verified) : undefined;
		const actuallyActivated = Number(!!activated);

		if (title && (title.length < 3 || title.length > 32)) {
			return fail(403, { errors: ['invalidTitle'] });
		}

		if (description && description.length > 1000) {
			return fail(403, { errors: ['invalidDescription'] });
		}

		await OAuth2Clients.update(details as OAuth2Client, {
			title,
			description,
			verified: actuallyVerified,
			activated: actuallyActivated
		});

		return { errors: [] };
	},
	delete: async ({ locals, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		if (details.activated === 1) {
			return fail(400, { errors: ['deleteActivated'] });
		}

		if (details.ownerId !== currentUser.id && !fullPrivileges) {
			return fail(403, { errors: ['forbidden'] });
		}

		await OAuth2Clients.deleteClient(details as OAuth2Client);

		return redirect(303, '/ssoadmin/oauth2');
	},
	regenerate: async ({ locals, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		if (!fullPrivileges && !details.isOwner) {
			return fail(403, { errors: ['forbidden'] });
		}

		await OAuth2Clients.update(details as OAuth2Client, {
			client_secret: CryptoUtils.generateSecret()
		});

		return { errors: [] };
	},
	removeUrl: async ({ locals, url, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const id = Number(url.searchParams.get('id'));
		if (isNaN(id)) {
			return fail(400, { errors: ['invalidUrlId'] });
		}

		await OAuth2Clients.deleteUrl(details as OAuth2Client, id);

		return { errors: [] };
	},
	addUrl: async ({ locals, request, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const body = await request.formData();
		const { type, url } = Changesets.take<AddUrlRequest>(['type', 'url'], body);
		if (!type || !OAuth2Clients.availableUrlTypes.includes(type)) {
			return fail(400, { errors: ['invalidUrlType'] });
		}

		if (!url) {
			return fail(400, { errors: ['invalidUrl'] });
		}

		await OAuth2Clients.addUrl(details as OAuth2Client, type, url);

		return { errors: [] };
	},
	removePrivilege: async ({ locals, url, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const id = Number(url.searchParams.get('id'));
		if (isNaN(id)) {
			return fail(400, { errors: ['invalidPrivilegeId'] });
		}

		await OAuth2Clients.deletePrivilege(details as OAuth2Client, id);

		return { errors: [] };
	},
	addPrivilege: async ({ locals, request, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const body = await request.formData();
		const { name } = Changesets.take<AddPrivilegeRequest>(['name'], body);

		if (!name || !privilegeRegex.test(name)) {
			return fail(400, { errors: ['invalidPrivilege'] });
		}

		await OAuth2Clients.addPrivilege(details as OAuth2Client, name);

		return { errors: [] };
	},
	grants: async ({ locals, request, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const allowedGrants = fullPrivileges
			? OAuth2Clients.availableGrantTypes
			: OAuth2Clients.userSetGrants;

		const body = await request.formData();
		const values = Array.from(body.keys());
		values.unshift('authorization_code');
		const deduplicatedAllowedGrants = values.filter(
			(value, index, array) => allowedGrants.includes(value) && array.indexOf(value) === index
		);

		await OAuth2Clients.update(details as OAuth2Client, {
			grants: deduplicatedAllowedGrants.join(' ')
		});

		return { errors: [] };
	},
	scopes: async ({ locals, request, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const allowedScopes = fullPrivileges
			? OAuth2Clients.availableScopes
			: OAuth2Clients.userSetScopes;

		const body = await request.formData();
		const values = Array.from(body.keys());
		values.unshift('profile');
		const deduplicatedAllowedScopes = values.filter(
			(value, index, array) => allowedScopes.includes(value) && array.indexOf(value) === index
		);

		await OAuth2Clients.update(details as OAuth2Client, {
			scope: OAuth2Clients.joinScope(deduplicatedAllowedScopes)
		});

		return { errors: [] };
	},
	avatar: async ({ request, locals, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const formData = Object.fromEntries(await request.formData());
		if (!(formData.file as File)?.name || (formData.file as File).name === 'undefined') {
			return fail(400, {
				errors: ['noFile']
			});
		}

		const { file } = formData as { file: File };

		await Uploads.saveClientAvatar(details as OAuth2Client, currentUser, file);

		return { errors: [] };
	},
	removeAvatar: async ({ locals, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		await Uploads.removeClientAvatar(details as OAuth2Client);

		return { errors: [] };
	},
	invite: async ({ locals, request, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const body = await request.formData();
		const { email } = Changesets.take<InviteRequest>(['email'], body);

		if (!email || !emailRegex.test(email)) {
			return fail(400, { errors: ['invalidEmail'] });
		}

		const managers = await OAuth2Clients.getManagers(details as OAuth2Client);
		if (managers.some((entry) => entry.email.toLowerCase() === email.toLowerCase())) {
			return fail(400, { errors: ['emailExists'] });
		}

		await OAuth2Clients.sendManagerInvitationEmail(details as OAuth2Client, currentUser, email);

		return { errors: [] };
	},
	removeManager: async ({ locals, url, params: { uuid } }) => {
		const { currentUser, userSession } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const fullPrivileges = hasPrivileges(userSession.privileges || [], ['admin:oauth2']);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: fullPrivileges,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const id = Number(url.searchParams.get('id'));
		if (isNaN(id)) {
			return fail(400, { errors: ['invalidManagerId'] });
		}

		await OAuth2Clients.removeManager(details as OAuth2Client, id);

		return { errors: [] };
	}
};

export const load = async ({ params: { uuid }, parent }) => {
	const { user } = await parent();
	const currentUser = await Users.getBySession(user);
	AdminUtils.checkPrivileges(user, [['admin:oauth2', 'self:oauth2']]);

	const fullPrivileges = hasPrivileges(user.privileges, ['admin:oauth2']);

	const {
		list: [details]
	} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
		clientId: uuid,
		listAll: fullPrivileges,
		omitSecret: false
	});

	if (!details) {
		return error(404, 'Client not found');
	}

	const privileges = await Users.getAvailablePrivileges(details.id);
	const users = await OAuth2Clients.getAuthorizedUsers(details as OAuth2Client);
	const managers = await OAuth2Clients.getManagers(details as OAuth2Client);

	return {
		users,
		managers,
		availableUrls: OAuth2Clients.availableUrlTypes,
		availablePrivileges: privileges,
		availableGrants: fullPrivileges
			? OAuth2Clients.availableGrantTypes
			: OAuth2Clients.userSetGrants,
		availableScopes: fullPrivileges ? OAuth2Clients.availableScopes : OAuth2Clients.userSetScopes,
		fullPrivileges,
		details
	};
};
