import { AdminUtils } from '$lib/server/admin-utils';
import { Changesets } from '$lib/server/changesets.js';
import type { OAuth2Client, User } from '$lib/server/drizzle';
import { OAuth2Clients } from '$lib/server/oauth2';
import { Users } from '$lib/server/users';
import { UsersAdmin } from '$lib/server/users/admin';
import { error, redirect } from '@sveltejs/kit';

interface PrivilegesRequest {
	privileges: string;
}

export const actions = {
	privileges: async ({ locals, params: { uuid, user: userId }, request }) => {
		const { currentUser } = await UsersAdmin.getActionUser(locals, [
			['admin:oauth2', 'self:oauth2']
		]);

		const {
			list: [details]
		} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
			clientId: uuid,
			listAll: false,
			omitSecret: false
		});

		if (!details) {
			return error(404, 'Client not found');
		}

		const [auth] = await OAuth2Clients.getAuthorizedUsers(details as OAuth2Client, userId);
		if (!auth) {
			return error(404, 'User not found');
		}

		const targetUser = await Users.getByUuid(userId);
		if (!targetUser) {
			return error(404, 'User not found');
		}

		const availablePrivileges = await Users.getAvailablePrivileges(details.id);
		const body = await request.formData();

		const { privileges } = Changesets.take<PrivilegesRequest>(['privileges'], body);
		const splitFilter = (privileges || '').split(',').reduce<number[]>((final, id) => {
			const privId = Number(id);

			if (
				isNaN(privId) ||
				final.includes(privId) ||
				!availablePrivileges.some((entry) => entry.id === privId)
			) {
				return final;
			}

			return [...final, privId];
		}, []);

		await Users.setUserPrivileges(targetUser, splitFilter, details.id);

		return redirect(303, '..');
	}
};

export const load = async ({ params: { uuid, user: userId }, parent }) => {
	const { user } = await parent();
	const currentUser = await Users.getBySession(user);
	AdminUtils.checkPrivileges(user, [['admin:oauth2', 'self:oauth2']]);

	const {
		list: [details]
	} = await OAuth2Clients.getClientByAdminUser(currentUser as User, {
		clientId: uuid,
		listAll: false,
		omitSecret: false
	});

	if (!details) {
		return error(404, 'Client not found');
	}

	const [auth] = await OAuth2Clients.getAuthorizedUsers(details as OAuth2Client, userId);
	if (!auth) {
		return error(404, 'User not found');
	}

	const availablePrivileges = await Users.getAvailablePrivileges(details.id);

	return {
		auth,
		details,
		availablePrivileges
	};
};
